@@ -1,30 +1,26 @@
 /* File I/O for GNU DIFF.
 
-   Copyright (C) 1988, 1989, 1992, 1993, 1994, 1995, 1998, 2001, 2002
+   Copyright (C) 1988-1989, 1992-1995, 1998, 2001-2002, 2004, 2006, 2009-2010
    Free Software Foundation, Inc.
 
    This file is part of GNU DIFF.
 
-   GNU DIFF is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+   the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
 
-   GNU DIFF is distributed in the hope that it will be useful,
+   This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; see the file COPYING.
-   If not, write to the Free Software Foundation,
-   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #include "diff.h"
 #include <cmpbuf.h>
 #include <file-type.h>
-#include <regex.h>
-#include <setmode.h>
 #include <xalloc.h>
 
 /* Rotate an unsigned value to the left.  */
@@ -35,7 +31,7 @@
 
 /* The type of a hash value.  */
 typedef size_t hash_value;
-verify (hash_value_is_unsigned, ! TYPE_SIGNED (hash_value));
+verify (! TYPE_SIGNED (hash_value));
 
 /* Lines are put into equivalence classes of lines that match in lines_differ.
    Each equivalence class is represented by one of these structures,
@@ -115,32 +111,17 @@ sip (struct file_data *current, bool skip_test)
 	{
 	  /* Check first part of file to see if it's a binary file.  */
 
-	  bool was_binary = set_binary_mode (current->desc, 1);
-	  off_t buffered;
+	  /* FIXME: if O_BINARY, this should revert to text mode
+	     if the file is not binary.  */
+
 	  file_block_read (current, current->bufsize);
-	  buffered = current->buffered;
-
-	  if (! was_binary)
-	    {
-	      /* Revert to text mode and seek back to the beginning to
-		 reread the file.  Use relative seek, since file
-		 descriptors like stdin might not start at offset
-		 zero.  */
-
-	      if (lseek (current->desc, - buffered, SEEK_CUR) == -1)
-		pfatal_with_name (current->name);
-	      set_binary_mode (current->desc, 0);
-	      current->buffered = 0;
-	      current->eof = 0;
-	    }
-
-	  return binary_file_p (current->buffer, buffered);
+	  return binary_file_p (current->buffer, current->buffered);
 	}
     }
 
   current->buffered = 0;
-  current->eof = 0;
-  return 0;
+  current->eof = false;
+  return false;
 }
 
 /* Slurp the rest of the current file completely into memory.  */
@@ -218,7 +199,7 @@ static void
 find_and_hash_each_line (struct file_data *current)
 {
   hash_value h;
-  unsigned char const *p = (unsigned char const *) current->prefix_end;
+  char const *p = current->prefix_end;
   unsigned char c;
   lin i, *bucket;
   size_t length;
@@ -239,9 +220,9 @@ find_and_hash_each_line (struct file_data *current)
   bool same_length_diff_contents_compare_anyway =
     diff_length_compare_anyway | ignore_case;
 
-  while ((char const *) p < suffix_begin)
+  while (p < suffix_begin)
     {
-      char const *ip = (char const *) p;
+      char const *ip = p;
 
       h = 0;
 
@@ -251,25 +232,25 @@ find_and_hash_each_line (struct file_data *current)
 	  {
 	  case IGNORE_ALL_SPACE:
 	    while ((c = *p++) != '\n')
-	      if (! ISSPACE (c))
-		h = HASH (h, TOLOWER (c));
+	      if (! isspace (c))
+		h = HASH (h, tolower (c));
 	    break;
 
 	  case IGNORE_SPACE_CHANGE:
 	    while ((c = *p++) != '\n')
 	      {
-		if (ISSPACE (c))
+		if (isspace (c))
 		  {
 		    do
 		      if ((c = *p++) == '\n')
 			goto hashing_done;
-		    while (ISSPACE (c));
+		    while (isspace (c));
 
 		    h = HASH (h, ' ');
 		  }
 
 		/* C is now the first non-space.  */
-		h = HASH (h, TOLOWER (c));
+		h = HASH (h, tolower (c));
 	      }
 	    break;
 
@@ -299,7 +280,7 @@ find_and_hash_each_line (struct file_data *current)
 		      break;
 
 		    default:
-		      c = TOLOWER (c);
+		      c = tolower (c);
 		      column++;
 		      break;
 		    }
@@ -313,7 +294,7 @@ find_and_hash_each_line (struct file_data *current)
 
 	  default:
 	    while ((c = *p++) != '\n')
-	      h = HASH (h, TOLOWER (c));
+	      h = HASH (h, tolower (c));
 	    break;
 	  }
       else
@@ -321,19 +302,19 @@ find_and_hash_each_line (struct file_data *current)
 	  {
 	  case IGNORE_ALL_SPACE:
 	    while ((c = *p++) != '\n')
-	      if (! ISSPACE (c))
+	      if (! isspace (c))
 		h = HASH (h, c);
 	    break;
 
 	  case IGNORE_SPACE_CHANGE:
 	    while ((c = *p++) != '\n')
 	      {
-		if (ISSPACE (c))
+		if (isspace (c))
 		  {
 		    do
 		      if ((c = *p++) == '\n')
 			goto hashing_done;
-		    while (ISSPACE (c));
+		    while (isspace (c));
 
 		    h = HASH (h, ' ');
 		  }
@@ -389,9 +370,9 @@ find_and_hash_each_line (struct file_data *current)
    hashing_done:;
 
       bucket = &buckets[h % nbuckets];
-      length = (char const *) p - ip - 1;
+      length = p - ip - 1;
 
-      if ((char const *) p == bufend
+      if (p == bufend
 	  && current->missing_newline
 	  && ROBUST_OUTPUT_STYLE (output_style))
 	{
@@ -402,7 +383,7 @@ find_and_hash_each_line (struct file_data *current)
 
 	  /* Omit the inserted newline when computing linbuf later.  */
 	  p--;
-	  bufend = suffix_begin = (char const *) p;
+	  bufend = suffix_begin = p;
 	}
 
       for (i = *bucket;  ;  i = eqs[i].next)
@@ -487,9 +468,9 @@ find_and_hash_each_line (struct file_data *current)
 			     (alloc_lines - linbuf_base) * sizeof *linbuf);
 	  linbuf -= linbuf_base;
 	}
-      linbuf[line] = (char const *) p;
+      linbuf[line] = p;
 
-      if ((char const *) p == bufend)
+      if (p == bufend)
 	break;
 
       if (context <= i && no_diff_means_no_output)
@@ -524,11 +505,11 @@ prepare_text (struct file_data *current)
   char *dst;
 
   if (buffered == 0 || p[buffered - 1] == '\n')
-    current->missing_newline = 0;
+    current->missing_newline = false;
   else
     {
       p[buffered++] = '\n';
-      current->missing_newline = 1;
+      current->missing_newline = true;
     }
 
   if (!p)
@@ -666,10 +647,11 @@ find_identical_ends (struct file_data filevec[])
       beg0 = filevec[0].prefix_end + (n0 < n1 ? 0 : n0 - n1);
 
       /* Scan back until chars don't match or we reach that point.  */
-      for (; p0 != beg0; p0--, p1--)
-	if (*p0 != *p1)
+      while (p0 != beg0)
+	if (*--p0 != *--p1)
 	  {
 	    /* Point at the first char of the matching suffix.  */
+	    ++p0, ++p1;
 	    beg0 = p0;
 	    break;
 	  }
@@ -797,8 +779,7 @@ static unsigned char const prime_offset[] =
 
 /* Verify that this host's size_t is not too wide for the above table.  */
 
-verify (enough_prime_offsets,
-	sizeof (size_t) * CHAR_BIT <= sizeof prime_offset);
+verify (sizeof (size_t) * CHAR_BIT <= sizeof prime_offset);
 
 /* Given a vector of two file_data objects, read the file associated
    with each one, and build the table of equivalence classes.
@@ -822,9 +803,8 @@ read_files (struct file_data filevec[], bool pretend_binary)
     }
   if (appears_binary)
     {
-      set_binary_mode (filevec[0].desc, 1);
-      set_binary_mode (filevec[1].desc, 1);
-      return 1;
+      /* FIXME: If O_BINARY, this should set both files to binary mode.  */
+      return true;
     }
 
   find_identical_ends (filevec);
@@ -856,5 +836,5 @@ read_files (struct file_data filevec[], bool pretend_binary)
   free (equivs);
   free (buckets - 1);
 
-  return 0;
+  return false;
 }
